// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    bundle::{BundleType, BundleValue},
    expr::{
        ops::{ArrayIndex, ArrayLiteral, ExprIndex},
        Expr, ToExpr,
    },
    intern::{Intern, Interned, InternedCompare, Memoize},
    module::{
        transform::visit::{Fold, Folder, Visit, Visitor},
        ModuleBuilder, NormalModule,
    },
    source_location::SourceLocation,
    ty::{
        CanonicalType, CanonicalTypeKind, CanonicalValue, Connect, DynCanonicalType,
        DynCanonicalValue, DynType, DynValueTrait, FixedType, MatchVariantWithoutScope, Type,
        TypeEnum, Value, ValueEnum,
    },
    util::{ConstBool, GenericConstBool, MakeMutSlice},
};
use bitvec::{slice::BitSlice, vec::BitVec};
use std::{
    any::Any,
    borrow::{Borrow, BorrowMut},
    fmt,
    hash::Hash,
    marker::PhantomData,
    ops::IndexMut,
    sync::Arc,
};

mod sealed {
    pub trait Sealed {}
}

pub trait ValueArrayOrSlice:
    sealed::Sealed
    + BorrowMut<[<Self as ValueArrayOrSlice>::Element]>
    + AsRef<[<Self as ValueArrayOrSlice>::Element]>
    + AsMut<[<Self as ValueArrayOrSlice>::Element]>
    + Hash
    + fmt::Debug
    + Eq
    + Send
    + Sync
    + 'static
    + IndexMut<usize, Output = <Self as ValueArrayOrSlice>::Element>
    + ToOwned
    + InternedCompare
{
    type Element: Value<Type = <Self as ValueArrayOrSlice>::ElementType>;
    type ElementType: Type<Value = <Self as ValueArrayOrSlice>::Element>;
    type LenType: 'static + Copy + Ord + fmt::Debug + Hash + Send + Sync;
    type Match: 'static
        + Clone
        + Eq
        + fmt::Debug
        + Hash
        + Send
        + Sync
        + BorrowMut<[Expr<Self::Element>]>;
    type MaskVA: ValueArrayOrSlice<
            Element = <Self::ElementType as Type>::MaskValue,
            ElementType = <Self::ElementType as Type>::MaskType,
            LenType = Self::LenType,
            MaskVA = Self::MaskVA,
        > + ?Sized;
    type IsFixedLen: GenericConstBool;
    const FIXED_LEN_TYPE: Option<Self::LenType>;
    fn make_match(array: Expr<Array<Self>>) -> Self::Match;
    fn len_from_len_type(v: Self::LenType) -> usize;
    #[allow(clippy::result_unit_err)]
    fn try_len_type_from_len(v: usize) -> Result<Self::LenType, ()>;
    fn len_type(&self) -> Self::LenType;
    fn len(&self) -> usize;
    fn is_empty(&self) -> bool;
    fn iter(&self) -> std::slice::Iter<Self::Element> {
        Borrow::<[_]>::borrow(self).iter()
    }
    fn clone_to_arc(&self) -> Arc<Self>;
    fn arc_make_mut(v: &mut Arc<Self>) -> &mut Self;
    fn arc_to_arc_slice(self: Arc<Self>) -> Arc<[Self::Element]>;
}

impl<T> sealed::Sealed for [T] {}

impl<V: Value> ValueArrayOrSlice for [V]
where
    V::Type: Type<Value = V>,
{
    type Element = V;
    type ElementType = V::Type;
    type LenType = usize;
    type Match = Box<[Expr<V>]>;
    type MaskVA = [<Self::ElementType as Type>::MaskValue];
    type IsFixedLen = ConstBool<false>;
    const FIXED_LEN_TYPE: Option<Self::LenType> = None;

    fn make_match(array: Expr<Array<Self>>) -> Self::Match {
        (0..array.canonical_type().len())
            .map(|index| ArrayIndex::<V::Type>::new_unchecked(array.canonical(), index).to_expr())
            .collect()
    }

    fn len_from_len_type(v: Self::LenType) -> usize {
        v
    }

    fn try_len_type_from_len(v: usize) -> Result<Self::LenType, ()> {
        Ok(v)
    }

    fn len_type(&self) -> Self::LenType {
        self.len()
    }

    fn len(&self) -> usize {
        <[_]>::len(self)
    }

    fn is_empty(&self) -> bool {
        <[_]>::is_empty(self)
    }

    fn clone_to_arc(&self) -> Arc<Self> {
        Arc::from(self)
    }

    fn arc_make_mut(v: &mut Arc<Self>) -> &mut Self {
        MakeMutSlice::make_mut_slice(v)
    }

    fn arc_to_arc_slice(self: Arc<Self>) -> Arc<[Self::Element]> {
        self
    }
}

impl<T, const N: usize> sealed::Sealed for [T; N] {}

impl<V: Value, const N: usize> ValueArrayOrSlice for [V; N]
where
    V::Type: Type<Value = V>,
{
    type Element = V;
    type ElementType = V::Type;
    type LenType = ();
    type Match = [Expr<V>; N];
    type MaskVA = [<Self::ElementType as Type>::MaskValue; N];
    type IsFixedLen = ConstBool<true>;
    const FIXED_LEN_TYPE: Option<Self::LenType> = Some(());

    fn make_match(array: Expr<Array<Self>>) -> Self::Match {
        std::array::from_fn(|index| {
            ArrayIndex::<V::Type>::new_unchecked(array.canonical(), index).to_expr()
        })
    }

    fn len_from_len_type(_v: Self::LenType) -> usize {
        N
    }

    fn try_len_type_from_len(v: usize) -> Result<Self::LenType, ()> {
        if v == N {
            Ok(())
        } else {
            Err(())
        }
    }

    fn len_type(&self) -> Self::LenType {}

    fn len(&self) -> usize {
        N
    }

    fn is_empty(&self) -> bool {
        N == 0
    }

    fn clone_to_arc(&self) -> Arc<Self> {
        Arc::new(self.clone())
    }

    fn arc_make_mut(v: &mut Arc<Self>) -> &mut Self {
        Arc::make_mut(v)
    }

    fn arc_to_arc_slice(self: Arc<Self>) -> Arc<[Self::Element]> {
        self
    }
}

#[derive(Debug, PartialEq, Eq, Hash)]
pub struct ArrayType<VA: ValueArrayOrSlice + ?Sized> {
    element: VA::ElementType,
    len: VA::LenType,
    bit_width: usize,
}

pub trait ArrayTypeTrait:
    Type<
        CanonicalType = ArrayType<[DynCanonicalValue]>,
        Value = Array<<Self as ArrayTypeTrait>::ValueArrayOrSlice>,
        CanonicalValue = Array<[DynCanonicalValue]>,
        MaskType = ArrayType<
            <<Self as ArrayTypeTrait>::ValueArrayOrSlice as ValueArrayOrSlice>::MaskVA,
        >,
    > + From<ArrayType<<Self as ArrayTypeTrait>::ValueArrayOrSlice>>
    + Into<ArrayType<<Self as ArrayTypeTrait>::ValueArrayOrSlice>>
    + BorrowMut<ArrayType<<Self as ArrayTypeTrait>::ValueArrayOrSlice>>
    + sealed::Sealed
    + Connect<Self>
{
    type ValueArrayOrSlice: ValueArrayOrSlice<Element = Self::Element, ElementType = Self::ElementType>
        + ?Sized;
    type Element: Value<Type = Self::ElementType>;
    type ElementType: Type<Value = Self::Element>;
}

impl<VA: ValueArrayOrSlice + ?Sized> sealed::Sealed for ArrayType<VA> {}

impl<VA: ValueArrayOrSlice + ?Sized> ArrayTypeTrait for ArrayType<VA> {
    type ValueArrayOrSlice = VA;
    type Element = VA::Element;
    type ElementType = VA::ElementType;
}

impl<VA: ValueArrayOrSlice + ?Sized> Clone for ArrayType<VA> {
    fn clone(&self) -> Self {
        Self {
            element: self.element.clone(),
            len: self.len,
            bit_width: self.bit_width,
        }
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> Copy for ArrayType<VA> where VA::ElementType: Copy {}

impl<VA: ValueArrayOrSlice + ?Sized> ArrayType<VA> {
    pub fn element(&self) -> &VA::ElementType {
        &self.element
    }
    pub fn len(&self) -> usize {
        VA::len_from_len_type(self.len)
    }
    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }
    pub fn bit_width(&self) -> usize {
        self.bit_width
    }
    pub fn into_slice_type(self) -> ArrayType<[VA::Element]> {
        ArrayType {
            len: self.len(),
            element: self.element,
            bit_width: self.bit_width,
        }
    }
    #[track_caller]
    pub fn new_with_len(element: VA::ElementType, len: usize) -> Self {
        Self::new_with_len_type(
            element,
            VA::try_len_type_from_len(len).expect("length should match"),
        )
    }
    #[track_caller]
    pub fn new_with_len_type(element: VA::ElementType, len: VA::LenType) -> Self {
        let Some(bit_width) = VA::len_from_len_type(len).checked_mul(element.bit_width()) else {
            panic!("array is too big: bit-width overflowed");
        };
        ArrayType {
            element,
            len,
            bit_width,
        }
    }
}

impl<VA: ValueArrayOrSlice + ?Sized, State: ?Sized + Folder> Fold<State> for ArrayType<VA>
where
    VA::ElementType: Fold<State>,
{
    fn fold(self, state: &mut State) -> Result<Self, State::Error> {
        state.fold_array_type(self)
    }
    fn default_fold(self, state: &mut State) -> Result<Self, State::Error> {
        Ok(Self::new_with_len_type(self.element.fold(state)?, self.len))
    }
}

impl<VA: ValueArrayOrSlice + ?Sized, State: ?Sized + Visitor> Visit<State> for ArrayType<VA>
where
    VA::ElementType: Visit<State>,
{
    fn visit(&self, state: &mut State) -> Result<(), State::Error> {
        state.visit_array_type(self)
    }
    fn default_visit(&self, state: &mut State) -> Result<(), State::Error> {
        self.element.visit(state)
    }
}

impl<V: Value, const N: usize> ArrayType<[V; N]>
where
    V::Type: Type<Value = V>,
{
    pub fn new_array(element: V::Type) -> Self {
        ArrayType::new_with_len_type(element, ())
    }
}

impl<V: Value, const N: usize> FixedType for ArrayType<[V; N]>
where
    V::Type: FixedType<Value = V>,
{
    fn fixed_type() -> Self {
        Self::new_array(FixedType::fixed_type())
    }
}

impl<V: Value> ArrayType<[V]>
where
    V::Type: Type<Value = V>,
{
    pub fn new_slice(element: V::Type, len: usize) -> Self {
        ArrayType::new_with_len_type(element, len)
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> Type for ArrayType<VA> {
    type CanonicalType = ArrayType<[DynCanonicalValue]>;
    type Value = Array<VA>;
    type CanonicalValue = Array<[DynCanonicalValue]>;
    type MaskType = ArrayType<VA::MaskVA>;
    type MaskValue = Array<VA::MaskVA>;
    type MatchVariant = VA::Match;
    type MatchActiveScope = ();
    type MatchVariantAndInactiveScope = MatchVariantWithoutScope<VA::Match>;
    type MatchVariantsIter = std::iter::Once<Self::MatchVariantAndInactiveScope>;

    fn match_variants<IO: BundleValue>(
        this: Expr<Self::Value>,
        module_builder: &mut ModuleBuilder<IO, NormalModule>,
        source_location: SourceLocation,
    ) -> Self::MatchVariantsIter
    where
        IO::Type: BundleType<Value = IO>,
    {
        let _ = module_builder;
        let _ = source_location;
        std::iter::once(MatchVariantWithoutScope(VA::make_match(this)))
    }

    fn mask_type(&self) -> Self::MaskType {
        #[derive(Clone, Hash, Eq, PartialEq)]
        struct ArrayMaskTypeMemoize<T: ArrayTypeTrait>(PhantomData<T>);
        impl<T: ArrayTypeTrait> Copy for ArrayMaskTypeMemoize<T> {}
        impl<T: ArrayTypeTrait> Memoize for ArrayMaskTypeMemoize<T> {
            type Input = ArrayType<T::ValueArrayOrSlice>;
            type InputOwned = ArrayType<T::ValueArrayOrSlice>;
            type Output = <ArrayType<T::ValueArrayOrSlice> as Type>::MaskType;

            fn inner(self, input: &Self::Input) -> Self::Output {
                ArrayType::new_with_len_type(input.element.mask_type(), input.len)
            }
        }
        ArrayMaskTypeMemoize::<Self>(PhantomData).get(self)
    }

    fn canonical(&self) -> Self::CanonicalType {
        ArrayType {
            element: self.element.canonical_dyn(),
            len: self.len(),
            bit_width: self.bit_width,
        }
    }

    fn source_location(&self) -> SourceLocation {
        SourceLocation::builtin()
    }

    fn type_enum(&self) -> TypeEnum {
        TypeEnum::ArrayType(self.canonical())
    }

    fn from_canonical_type(t: Self::CanonicalType) -> Self {
        Self {
            element: VA::ElementType::from_dyn_canonical_type(t.element),
            len: VA::try_len_type_from_len(t.len).expect("length should match"),
            bit_width: t.bit_width,
        }
    }

    fn as_dyn_canonical_type_impl(this: &Self) -> Option<&dyn DynCanonicalType> {
        Some(<dyn Any>::downcast_ref::<ArrayType<[DynCanonicalValue]>>(
            this,
        )?)
    }
}

impl<Lhs: ValueArrayOrSlice + ?Sized, Rhs: ValueArrayOrSlice + ?Sized> Connect<ArrayType<Rhs>>
    for ArrayType<Lhs>
{
}

impl CanonicalType for ArrayType<[DynCanonicalValue]> {
    const CANONICAL_TYPE_KIND: CanonicalTypeKind = CanonicalTypeKind::ArrayType;
}

#[derive(Debug, PartialEq, Eq, Hash)]
pub struct Array<VA: ValueArrayOrSlice + ?Sized> {
    element_ty: VA::ElementType,
    value: Arc<VA>,
}

impl<VA: ValueArrayOrSlice + ?Sized> Clone for Array<VA> {
    fn clone(&self) -> Self {
        Self {
            element_ty: self.element_ty.clone(),
            value: self.value.clone(),
        }
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> ToExpr for Array<VA> {
    type Type = ArrayType<VA>;

    fn ty(&self) -> Self::Type {
        ArrayType::new_with_len_type(self.element_ty.clone(), self.value.len_type())
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        Expr::from_value(self)
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> Value for Array<VA> {
    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue {
        Array {
            element_ty: self.element_ty.canonical_dyn(),
            value: AsRef::<[_]>::as_ref(&*self.value)
                .iter()
                .map(|v| v.to_canonical_dyn())
                .collect(),
        }
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        #[derive(Hash, Eq, PartialEq)]
        struct ArrayToBitsMemoize<VA: ValueArrayOrSlice + ?Sized>(PhantomData<VA>);
        impl<VA: ValueArrayOrSlice + ?Sized> Clone for ArrayToBitsMemoize<VA> {
            fn clone(&self) -> Self {
                *self
            }
        }
        impl<VA: ValueArrayOrSlice + ?Sized> Copy for ArrayToBitsMemoize<VA> {}
        impl<VA: ValueArrayOrSlice + ?Sized> Memoize for ArrayToBitsMemoize<VA> {
            type Input = Array<VA>;
            type InputOwned = Array<VA>;
            type Output = Interned<BitSlice>;

            fn inner(self, input: &Self::Input) -> Self::Output {
                let mut bits = BitVec::with_capacity(input.ty().bit_width());
                for element in AsRef::<[_]>::as_ref(&*input.value).iter() {
                    bits.extend_from_bitslice(&element.to_bits());
                }
                Intern::intern_owned(bits)
            }
        }
        ArrayToBitsMemoize::<VA>(PhantomData).get(this)
    }
}

impl CanonicalValue for Array<[DynCanonicalValue]> {
    fn value_enum_impl(this: &Self) -> ValueEnum {
        ValueEnum::Array(this.clone())
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        Value::to_bits_impl(this)
    }
}

impl<VA: ValueArrayOrSlice + ?Sized> Array<VA> {
    pub fn element_ty(&self) -> &VA::ElementType {
        &self.element_ty
    }
    pub fn len(&self) -> usize {
        VA::len_from_len_type(self.value.len_type())
    }
    pub fn is_empty(&self) -> bool {
        self.len() == 0
    }
    pub fn value(&self) -> &Arc<VA> {
        &self.value
    }
    pub fn set_element(&mut self, index: usize, element: VA::Element) {
        assert_eq!(self.element_ty, element.ty());
        VA::arc_make_mut(&mut self.value)[index] = element;
    }
    pub fn new(element_ty: VA::ElementType, value: Arc<VA>) -> Self {
        for element in value.iter() {
            assert_eq!(element_ty, element.ty());
        }
        Self { element_ty, value }
    }
    pub fn into_slice(self) -> Array<[VA::Element]> {
        Array {
            element_ty: self.element_ty,
            value: self.value.arc_to_arc_slice(),
        }
    }
}

impl<VA: ValueArrayOrSlice + ?Sized, T: Into<Arc<VA>>> From<T> for Array<VA>
where
    VA::ElementType: FixedType,
{
    fn from(value: T) -> Self {
        Self::new(FixedType::fixed_type(), value.into())
    }
}

impl<E: ToExpr<Type = T>, T: FixedType> ToExpr for [E] {
    type Type = ArrayType<[T::Value]>;

    fn ty(&self) -> Self::Type {
        ArrayType::new_with_len_type(FixedType::fixed_type(), self.len())
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        let elements = Intern::intern_owned(Vec::from_iter(
            self.iter().map(|v| v.to_expr().to_canonical_dyn()),
        ));
        ArrayLiteral::new_unchecked(elements, self.ty()).to_expr()
    }
}

impl<E: ToExpr<Type = T>, T: FixedType> ToExpr for Vec<E> {
    type Type = ArrayType<[T::Value]>;

    fn ty(&self) -> Self::Type {
        <[E]>::ty(self)
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        <[E]>::to_expr(self)
    }
}

impl<E: ToExpr<Type = T>, T: FixedType> ToExpr for [E; 0] {
    type Type = ArrayType<[T::Value; 0]>;

    fn ty(&self) -> Self::Type {
        ArrayType::new_with_len_type(FixedType::fixed_type(), ())
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        Array::new(FixedType::fixed_type(), Arc::new([])).to_expr()
    }
}

macro_rules! impl_to_expr_for_non_empty_array {
    ($N:literal) => {
        impl<E: ToExpr<Type = T>, T: Type> ToExpr for [E; $N] {
            type Type = ArrayType<[T::Value; $N]>;

            fn ty(&self) -> Self::Type {
                ArrayType::new_with_len_type(self[0].ty(), ())
            }

            fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
                let elements = Intern::intern_owned(Vec::from_iter(
                    self.iter().map(|v| v.to_expr().to_canonical_dyn()),
                ));
                ArrayLiteral::new_unchecked(elements, self.ty()).to_expr()
            }
        }
    };
}

impl_to_expr_for_non_empty_array!(1);
impl_to_expr_for_non_empty_array!(2);
impl_to_expr_for_non_empty_array!(3);
impl_to_expr_for_non_empty_array!(4);
impl_to_expr_for_non_empty_array!(5);
impl_to_expr_for_non_empty_array!(6);
impl_to_expr_for_non_empty_array!(7);
impl_to_expr_for_non_empty_array!(8);
impl_to_expr_for_non_empty_array!(9);
impl_to_expr_for_non_empty_array!(10);
impl_to_expr_for_non_empty_array!(11);
impl_to_expr_for_non_empty_array!(12);
impl_to_expr_for_non_empty_array!(13);
impl_to_expr_for_non_empty_array!(14);
impl_to_expr_for_non_empty_array!(15);
impl_to_expr_for_non_empty_array!(16);
impl_to_expr_for_non_empty_array!(17);
impl_to_expr_for_non_empty_array!(18);
impl_to_expr_for_non_empty_array!(19);
impl_to_expr_for_non_empty_array!(20);
impl_to_expr_for_non_empty_array!(21);
impl_to_expr_for_non_empty_array!(22);
impl_to_expr_for_non_empty_array!(23);
impl_to_expr_for_non_empty_array!(24);
impl_to_expr_for_non_empty_array!(25);
impl_to_expr_for_non_empty_array!(26);
impl_to_expr_for_non_empty_array!(27);
impl_to_expr_for_non_empty_array!(28);
impl_to_expr_for_non_empty_array!(29);
impl_to_expr_for_non_empty_array!(30);
impl_to_expr_for_non_empty_array!(31);
impl_to_expr_for_non_empty_array!(32);

#[derive(Clone, Debug)]
pub struct ArrayIntoIter<VA: ValueArrayOrSlice> {
    array: Arc<VA>,
    indexes: std::ops::Range<usize>,
}

impl<VA: ValueArrayOrSlice> Iterator for ArrayIntoIter<VA> {
    type Item = VA::Element;

    fn next(&mut self) -> Option<Self::Item> {
        Some(self.array[self.indexes.next()?].clone())
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        self.indexes.size_hint()
    }
}

impl<VA: ValueArrayOrSlice> std::iter::FusedIterator for ArrayIntoIter<VA> {}

impl<VA: ValueArrayOrSlice> ExactSizeIterator for ArrayIntoIter<VA> {}

impl<VA: ValueArrayOrSlice> DoubleEndedIterator for ArrayIntoIter<VA> {
    fn next_back(&mut self) -> Option<Self::Item> {
        Some(self.array[self.indexes.next_back()?].clone())
    }
}

impl<VA: ValueArrayOrSlice> Array<VA> {
    pub fn iter(&self) -> std::slice::Iter<'_, VA::Element> {
        self.value.iter()
    }
}

impl<'a, VA: ValueArrayOrSlice> IntoIterator for &'a Array<VA> {
    type Item = &'a VA::Element;
    type IntoIter = std::slice::Iter<'a, VA::Element>;

    fn into_iter(self) -> Self::IntoIter {
        self.value.iter()
    }
}

impl<VA: ValueArrayOrSlice> IntoIterator for Array<VA> {
    type Item = VA::Element;
    type IntoIter = ArrayIntoIter<VA>;

    fn into_iter(self) -> Self::IntoIter {
        ArrayIntoIter {
            indexes: 0..self.len(),
            array: self.value,
        }
    }
}

#[derive(Clone, Debug)]
pub struct ArrayExprIter<VA: ValueArrayOrSlice> {
    array: Expr<Array<VA>>,
    indexes: std::ops::Range<usize>,
}

impl<VA: ValueArrayOrSlice> Iterator for ArrayExprIter<VA> {
    type Item = Expr<VA::Element>;

    fn next(&mut self) -> Option<Self::Item> {
        Some(ExprIndex::expr_index(self.array, self.indexes.next()?))
    }

    fn size_hint(&self) -> (usize, Option<usize>) {
        self.indexes.size_hint()
    }
}

impl<VA: ValueArrayOrSlice> std::iter::FusedIterator for ArrayExprIter<VA> {}

impl<VA: ValueArrayOrSlice> ExactSizeIterator for ArrayExprIter<VA> {}

impl<VA: ValueArrayOrSlice> DoubleEndedIterator for ArrayExprIter<VA> {
    fn next_back(&mut self) -> Option<Self::Item> {
        Some(ExprIndex::expr_index(self.array, self.indexes.next_back()?))
    }
}

impl<VA: ValueArrayOrSlice> IntoIterator for Expr<Array<VA>> {
    type Item = Expr<VA::Element>;
    type IntoIter = ArrayExprIter<VA>;

    fn into_iter(self) -> Self::IntoIter {
        self.iter()
    }
}

impl<VA: ValueArrayOrSlice> IntoIterator for &'_ Expr<Array<VA>> {
    type Item = Expr<VA::Element>;
    type IntoIter = ArrayExprIter<VA>;

    fn into_iter(self) -> Self::IntoIter {
        self.iter()
    }
}

impl<VA: ValueArrayOrSlice> Expr<Array<VA>> {
    pub fn len(self) -> usize {
        self.canonical_type().len()
    }
    pub fn is_empty(self) -> bool {
        self.canonical_type().is_empty()
    }
    pub fn iter(self) -> ArrayExprIter<VA> {
        ArrayExprIter {
            indexes: 0..self.len(),
            array: self,
        }
    }
}
