// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    clock::ClockDomain,
    expr::{Expr, ExprTrait, Flow, ToExpr},
    intern::Interned,
    module::{NameId, ScopedNameId},
    source_location::SourceLocation,
    ty::{DynCanonicalType, DynType, Type},
};
use std::fmt;

#[derive(Clone, Eq, PartialEq, Hash)]
pub struct Reg<T: Type> {
    name: ScopedNameId,
    source_location: SourceLocation,
    ty: T,
    clock_domain: Expr<ClockDomain>,
    init: Option<Expr<T::Value>>,
}

impl<T: Type + fmt::Debug> fmt::Debug for Reg<T> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            name,
            source_location: _,
            ty,
            clock_domain,
            init,
        } = self;
        f.debug_struct("Reg")
            .field("name", name)
            .field("ty", ty)
            .field("clock_domain", clock_domain)
            .field("init", init)
            .finish_non_exhaustive()
    }
}

impl<T: Type> ToExpr for Reg<T> {
    type Type = T;

    fn ty(&self) -> Self::Type {
        self.ty.clone()
    }

    fn to_expr(&self) -> Expr<<Self::Type as Type>::Value> {
        Expr::new_unchecked(self.expr_enum())
    }
}

impl<T: Type> Reg<T> {
    pub fn canonical(&self) -> Reg<T::CanonicalType> {
        let Self {
            name,
            source_location,
            ref ty,
            clock_domain,
            init,
        } = *self;
        Reg {
            name,
            source_location,
            ty: ty.canonical(),
            clock_domain,
            init: init.map(Expr::canonical),
        }
    }
    pub fn to_dyn_reg(&self) -> Reg<Interned<dyn DynType>> {
        let Self {
            name,
            source_location,
            ref ty,
            clock_domain,
            init,
        } = *self;
        Reg {
            name,
            source_location,
            ty: ty.to_dyn(),
            clock_domain,
            init: init.map(Expr::to_dyn),
        }
    }
    pub fn to_dyn_canonical_reg(&self) -> Reg<Interned<dyn DynCanonicalType>> {
        let Self {
            name,
            source_location,
            ref ty,
            clock_domain,
            init,
        } = *self;
        Reg {
            name,
            source_location,
            ty: ty.canonical_dyn(),
            clock_domain,
            init: init.map(Expr::to_canonical_dyn),
        }
    }
    #[track_caller]
    pub fn new_unchecked(
        scoped_name: ScopedNameId,
        source_location: SourceLocation,
        ty: T,
        clock_domain: Expr<ClockDomain>,
        init: Option<Expr<T::Value>>,
    ) -> Self {
        assert!(ty.is_storable(), "register type must be a storable type");
        if let Some(init) = init {
            assert_eq!(ty, init.ty(), "register's type must match init type");
        }
        Self {
            name: scoped_name,
            source_location,
            ty,
            clock_domain,
            init,
        }
    }
    pub fn source_location(&self) -> SourceLocation {
        self.source_location
    }
    pub fn containing_module_name(&self) -> Interned<str> {
        self.containing_module_name_id().0
    }
    pub fn containing_module_name_id(&self) -> NameId {
        self.name.0
    }
    pub fn name(&self) -> Interned<str> {
        self.name_id().0
    }
    pub fn name_id(&self) -> NameId {
        self.name.1
    }
    pub fn scoped_name(&self) -> ScopedNameId {
        self.name
    }
    pub fn clock_domain(&self) -> Expr<ClockDomain> {
        self.clock_domain
    }
    pub fn init(&self) -> Option<Expr<T::Value>> {
        self.init
    }
    pub fn flow(&self) -> Flow {
        Flow::Duplex
    }
    pub fn must_connect_to(&self) -> bool {
        false
    }
}
