// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    expr::{Expr, ToExpr},
    int::{UInt, UIntType},
    intern::Interned,
    source_location::SourceLocation,
    ty::{
        impl_match_values_as_self, CanonicalType, CanonicalTypeKind, CanonicalValue, Connect,
        DynCanonicalType, FixedType, Type, TypeEnum, Value, ValueEnum,
    },
    util::interned_bit,
};
use bitvec::slice::BitSlice;

pub trait ResetTypeTrait: CanonicalType + FixedType {}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug, Default)]
pub struct AsyncResetType;

impl AsyncResetType {
    pub const fn new() -> Self {
        Self
    }
}

impl Type for AsyncResetType {
    type Value = AsyncReset;
    type CanonicalType = AsyncResetType;
    type CanonicalValue = AsyncReset;
    type MaskType = UIntType<1>;
    type MaskValue = UInt<1>;

    impl_match_values_as_self!();

    fn mask_type(&self) -> Self::MaskType {
        UIntType::new()
    }

    fn canonical(&self) -> Self::CanonicalType {
        *self
    }

    fn source_location(&self) -> SourceLocation {
        SourceLocation::builtin()
    }

    fn type_enum(&self) -> TypeEnum {
        TypeEnum::AsyncReset(*self)
    }

    fn from_canonical_type(t: Self::CanonicalType) -> Self {
        t
    }

    fn as_dyn_canonical_type_impl(this: &Self) -> Option<&dyn DynCanonicalType> {
        Some(this)
    }
}

impl Connect<Self> for AsyncResetType {}

impl CanonicalType for AsyncResetType {
    const CANONICAL_TYPE_KIND: CanonicalTypeKind = CanonicalTypeKind::AsyncReset;
}

impl FixedType for AsyncResetType {
    fn fixed_type() -> Self {
        Self
    }
}

impl ResetTypeTrait for AsyncResetType {}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug)]
pub struct AsyncReset(pub bool);

impl ToExpr for AsyncReset {
    type Type = AsyncResetType;

    fn ty(&self) -> Self::Type {
        AsyncResetType
    }

    fn to_expr(&self) -> Expr<Self> {
        Expr::from_value(self)
    }
}

impl Value for AsyncReset {
    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue {
        *self
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        interned_bit(this.0)
    }
}

impl CanonicalValue for AsyncReset {
    fn value_enum_impl(this: &Self) -> ValueEnum {
        ValueEnum::AsyncReset(*this)
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        interned_bit(this.0)
    }
}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug, Default)]
pub struct SyncResetType;

impl SyncResetType {
    pub const fn new() -> Self {
        Self
    }
}

impl Type for SyncResetType {
    type CanonicalType = SyncResetType;
    type Value = SyncReset;
    type CanonicalValue = SyncReset;
    type MaskType = UIntType<1>;
    type MaskValue = UInt<1>;

    impl_match_values_as_self!();

    fn mask_type(&self) -> Self::MaskType {
        UIntType::new()
    }

    fn canonical(&self) -> Self::CanonicalType {
        *self
    }

    fn source_location(&self) -> SourceLocation {
        SourceLocation::builtin()
    }

    fn type_enum(&self) -> TypeEnum {
        TypeEnum::SyncReset(*self)
    }

    fn from_canonical_type(t: Self::CanonicalType) -> Self {
        t
    }

    fn as_dyn_canonical_type_impl(this: &Self) -> Option<&dyn DynCanonicalType> {
        Some(this)
    }
}

impl Connect<Self> for SyncResetType {}

impl CanonicalType for SyncResetType {
    const CANONICAL_TYPE_KIND: CanonicalTypeKind = CanonicalTypeKind::SyncReset;
}

impl FixedType for SyncResetType {
    fn fixed_type() -> Self {
        Self
    }
}

impl ResetTypeTrait for SyncResetType {}

#[derive(Copy, Clone, Eq, PartialEq, Hash, Debug, Default)]
pub struct SyncReset(pub bool);

impl ToExpr for SyncReset {
    type Type = SyncResetType;

    fn ty(&self) -> Self::Type {
        SyncResetType
    }

    fn to_expr(&self) -> Expr<Self> {
        Expr::from_value(self)
    }
}

impl Value for SyncReset {
    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue {
        *self
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        interned_bit(this.0)
    }
}

impl CanonicalValue for SyncReset {
    fn value_enum_impl(this: &Self) -> ValueEnum {
        ValueEnum::SyncReset(*this)
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        interned_bit(this.0)
    }
}

#[derive(Copy, Clone, Debug, Eq, PartialEq, Hash, Default)]
pub struct ResetType;

impl ResetType {
    pub const fn new() -> Self {
        Self
    }
}

impl Type for ResetType {
    type Value = Reset;
    type CanonicalType = ResetType;
    type CanonicalValue = Reset;
    type MaskType = UIntType<1>;
    type MaskValue = UInt<1>;

    impl_match_values_as_self!();

    fn mask_type(&self) -> Self::MaskType {
        UIntType::new()
    }

    fn canonical(&self) -> Self::CanonicalType {
        *self
    }

    fn source_location(&self) -> SourceLocation {
        SourceLocation::builtin()
    }

    fn type_enum(&self) -> TypeEnum {
        TypeEnum::Reset(*self)
    }

    fn from_canonical_type(t: Self::CanonicalType) -> Self {
        t
    }

    fn as_dyn_canonical_type_impl(this: &Self) -> Option<&dyn DynCanonicalType> {
        Some(this)
    }
}

impl Connect<Self> for ResetType {}

impl CanonicalType for ResetType {
    const CANONICAL_TYPE_KIND: CanonicalTypeKind = CanonicalTypeKind::Reset;
}

impl FixedType for ResetType {
    fn fixed_type() -> Self {
        Self
    }
}

impl ResetTypeTrait for ResetType {}

#[derive(Copy, Clone, Debug, Eq, PartialEq, Hash)]
pub enum Reset {}

impl ToExpr for Reset {
    type Type = ResetType;

    fn ty(&self) -> Self::Type {
        match *self {}
    }

    fn to_expr(&self) -> Expr<Self> {
        Expr::from_value(self)
    }
}

impl Value for Reset {
    fn to_canonical(&self) -> <Self::Type as Type>::CanonicalValue {
        *self
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        match *this {}
    }
}

impl CanonicalValue for Reset {
    fn value_enum_impl(this: &Self) -> ValueEnum {
        ValueEnum::Reset(*this)
    }
    fn to_bits_impl(this: &Self) -> Interned<BitSlice> {
        match *this {}
    }
}

macro_rules! make_to_reset {
    (
        $(#[from_value($from_value_ty:ty)])*
        $vis:vis trait $Trait:ident {
            fn $fn:ident(&self) -> Expr<$T:ty>;
        }
    ) => {
        $vis trait $Trait {
            fn $fn(&self) -> Expr<$T>;
        }

        impl<T: ?Sized + $Trait> $Trait for &'_ T {
            fn $fn(&self) -> Expr<$T> {
                (**self).$fn()
            }
        }

        impl<T: ?Sized + $Trait> $Trait for &'_ mut T {
            fn $fn(&self) -> Expr<$T> {
                (**self).$fn()
            }
        }

        impl<T: ?Sized + $Trait> $Trait for Box<T> {
            fn $fn(&self) -> Expr<$T> {
                (**self).$fn()
            }
        }

        impl $Trait for Expr<$T> {
            fn $fn(&self) -> Expr<$T> {
                *self
            }
        }

        impl $Trait for $T {
            fn $fn(&self) -> Expr<$T> {
                self.to_expr()
            }
        }

        $(impl $Trait for $from_value_ty {
            fn $fn(&self) -> Expr<$T> {
                self.to_expr().$fn()
            }
        })*
    };
}

make_to_reset! {
    #[from_value(SyncReset)]
    #[from_value(AsyncReset)]
    pub trait ToReset {
        fn to_reset(&self) -> Expr<Reset>;
    }
}

make_to_reset! {
    #[from_value(bool)]
    #[from_value(UInt<1>)]
    pub trait ToAsyncReset {
        fn to_async_reset(&self) -> Expr<AsyncReset>;
    }
}

make_to_reset! {
    #[from_value(bool)]
    #[from_value(UInt<1>)]
    pub trait ToSyncReset {
        fn to_sync_reset(&self) -> Expr<SyncReset>;
    }
}
